from . import command

cur_row = 0
cur_col = 0

def clear():
    """Clear the RGB-LCD display."""

    command("Clearing RGB LCD display", "Rc", "Rc-OK")


def print(message, row = 0, col = 0):
    """Print a message to the RGB LCD screen."

    Parameters:
      message (str): The message to print
      row (int): 0-1
      col (int): 0-15
    """

    global cur_row
    global cur_col
    
    encoded = message.encode('utf-8')

    command(f"Printing '{message}' to RGB LCD display @ ({row}, {col})",
            "Rp", "Rp-OK", bytes([row, col, len(encoded)]), encoded)

    cur_row = row
    cur_col = col + len(encoded)


def set_colour(red, green, blue):
    """Set the backlight colour of the RGB LCD screen.

    Parameters:
      red (int): 0-255
      green (int): 0-255
      blue (int): 0-255
    """

    command(f"Setting RGB LCD backlight colour to ({red}, {green}, {blue})",
            "RC", "RC-OK", bytes([red, green, blue]))



from engi1020.arduino.api import *
from time import sleep

def aToc(a):         
  m = 255/2   
  if -1 <= a <= 1:
   return (m*(a + 1))
  elif a < -1:
   return (0.0)       
  else:
   return 255

pressure_getTemp()
pressure_getPressure()
pressure_getAltitude()

digital_read(6)
digital_write(4, True)
analog_read(6)

for i in range(10):
    #buzzer_note(5, analog_read(0)+200, 1)
    #temp_humid_getTemp(3)
    #analog_read(6)
    #sleep(1)
    three_axis_getAccelX()
    #sleep(1)
    three_axis_getAccelY()
    #sleep(1)
    three_axis_getAccelZ()
    pressure_getTemp()
    pressure_getPressure()
    pressure_getAltitude()


while True:
  r = int(aToc(three_axis_getAccelX()))
  g = int(aToc(three_axis_getAccelY()))
  b = int(aToc(three_axis_getAccelZ()))
  rgb_lcd_print(f"r:{r},g:{g},b:{b}")
  rgb_lcd_colour(r, g, b)
  sleep(0.5) 


def plain(*values, action=False):
    """Trace values with plain black and white."""

    for v in values:
        print(v, end=" ")

    print()


def pretty(*values, action=False):
    """Trace values with pretty colours and xxd-style hexdump."""

    import hexdump
    from colorama import Fore, Style

    for v in values:
        if action:
            print(f"{Fore.YELLOW}{v}", end=" ")
        elif type(v) == str:
            print(f"{Fore.MAGENTA}{v}", end=" ")
        elif type(v) == int:
            print(f"{Fore.GREEN}{v}", end=" ")
        elif type(v) == bytes:
            print(Fore.CYAN)
            hexdump.hexdump(v)
        else:
            print(f"{Fore.BLUE}{v}", end=" ")

    print(Style.RESET_ALL)


def sparse(*values, action=False):
    """Trace actions only."""

    if not action:
        return

    plain(*values)


def null(*args, **kwargs):
    pass


def set_default(name):
    global default

    if name is None:
        default = null

    elif name.lower() == "plain":
        default = plain

    elif name.lower() == "pretty":
        default = pretty

    elif name.lower() == "sparse":
        default = sparse

    else:
        raise ValueError(
            f"Invalid tracer '{name}': expected plain, pretty or sparse")


# Default tracer
default = null



from . import command


def getCentimeters(pin):
    """Get distance from ultrasonic in cm."""

    return (command("Getting distance in cm", "uc", int, bytes([pin])))/100


def getInches(pin):
    """Get distance from ultrasonic in inches."""

    return (command("Getting distance in inches", "ui", int, bytes([pin])))/100



from . import command


def analog_read(pin):
    """Read a value from an analog port.

    Parameters:
      pin (int):     Analog pin to read from (e.g., 2 for A2).

    Returns:         Quantized analog value (0–1023).
    """

    return command(f"Reading from analog port {pin}", "ar", int, bytes([pin]))


def analog_write(pin, duty):
    """Write a value to an analog port.

    Parameters:
      pin (int):     Analog pin to write to (e.g., 2 for A2).
      duty (int):    Duty cycle to set (0–255).
    """

    command(f"Writing {duty} to analog port {pin}",
            "aw", "AW-OK", bytes([pin, duty]))


def digital_read(pin):
    """Read a value from a digital port.

    Parameters:
      pin (int):     Digital pin to read from (e.g., 4 for D4).

    Returns:         True or False.
    """

    return command(f"Reading from digital port {pin}", "dr", bool, bytes([pin]))


def digital_write(pin, value):
    """Write a value to a digital port.

    Parameters:
      pin (int):     Digital pin to write to (e.g., 4 for D4).
      value (bool):  Value to write (True or False).
    """

    command(f"Writing {value} to digital port {pin}",
            "dw", "DW-OK", bytes([pin, value]))

def buzzer_frequency(pin, freq):
    """Play a frequency on a buzzer in a given port

    Parameters:
      pin (int):     Pin to write to (e.g., 4 for D4).
      freq (int):    Note to play in Hz.
    """
    
    freq_bytes = freq.to_bytes(2, byteorder = "little")
    command(f"Playing {freq}Hz on buzzer in port {pin}",
             "bf", "BF-OK", bytes([pin, freq_bytes[0], freq_bytes[1]]))

def buzzer_note(pin, freq, sec):
    """Play a frequency on a buzzer in a given port

    Parameters:
      pin (int):     Pin to write to (e.g., 4 for D4).
      freq (int):    Note to play in Hz.
      sec (int/float):  Duration of note in seconds
    """
    freq_bytes = freq.to_bytes(2, byteorder = "little")
    sec *= 1000 #Convert to ms
    sec = int(sec)
    sec_bytes = sec.to_bytes(4, byteorder = "little")
    command(f"Playing {freq}Hz on buzzer in port {pin} for {sec} seconds",
             "bt", "BT-OK", bytes([pin, freq_bytes[0], freq_bytes[1], sec_bytes[0], sec_bytes[1], sec_bytes[2], sec_bytes[3]]))

def buzzer_stop(pin):
    """Play a frequency on a buzzer in a given port

    Parameters:
      pin (int):     Pin to write to (e.g., 4 for D4).
      freq (int):    Note to play in Hz.
    """
    
    command(f"Stopping buzzer in port {pin}",
             "bs", "BS-OK", bytes([pin]))





from . import command


# Cursor location
row = 0
col = 0

# Sizing multipliers
scale = 1


def clear():
    """Clear the OLED display."""

    global row, col

    command("Clearing OLED display", "oc", "OC-OK")
    row = 0
    col = 0


def print_message(message):
    """Print a message to the OLED display at the current cursor position."""

    global row, col

    message = str(message)
    
    while len(message) > 0:
        space = 16 - col

        next_chunk = message[:space]
        remainder = message[space:]

        while len(next_chunk) < 16:
            next_chunk += ' '

        encoded = next_chunk.encode('utf-8')

        command(f"Printing {next_chunk} at ({row}, {col})", "op", "OP-OK",
                bytes([row, col, len(encoded)]), encoded)

        row += 1*scale
        col = 0
        message = remainder

def move_cursor(new_row, new_col):
    global row, col

    row = new_row
    col = new_col
    



"""Student-visible API for Arduino programming in labs.

This module re-exports selected functions and values from other modules to hide
some of the overall module's complexity. It can be imported with:

from engi1020.arduino.api import *

"""

from .io import analog_read, analog_write, digital_read, digital_write, buzzer_frequency, buzzer_note, buzzer_stop
from .oled import clear as oled_clear, print_message as oled_print, move_cursor as oled_move_cursor
from .rgb_lcd import (
        clear as rgb_lcd_clear,
        set_colour as rgb_lcd_colour,
        print as rgb_lcd_print,

)

from .dht11 import (
        getTemp as temp_humid_get_temp,
        getHumidity as temp_humid_get_humidity
)

from .three_axis_accel import (
        getAccelX as three_axis_get_accelX,
        getAccelY as three_axis_get_accelY,
        getAccelZ as three_axis_get_accelZ,
)

from .pressure import (
        getTemperature as pressure_get_temp,
        getPressure as pressure_get_pressure,
        getAltitude as pressure_get_altitude,
)

from .servo import (
        setAngle as servo_set_angle,
        getAngle as servo_get_angle,
)

from .ultra_distance import (
        getCentimeters as ultra_get_centimeters,
        getInches as ultra_get_inches,
)



from . import command


## Need to cast to signed int
def getAccelX():
    """Get acceleration in x-axis from Grove Beginner Accelerometer."""
    result = command("Getting acceleration in x-axis", "3x", int)
    if result >= 2**15:
        result -= 2**16
    return (result/100)

def getAccelY():
    """Get acceleration in x-axis from Grove Beginner Accelerometer."""

    result = command("Getting acceleration in y-axis", "3y", int)
    if result >= 2**15:
        result -= 2**16
    return (result/100)


def getAccelZ():
    """Get acceleration in z-axis from Grove Beginner Accelerometer."""

    result = command("Getting acceleration in z-axis", "3z", int)
    if result >= 2**15:
        result -= 2**16
    return (result/100)




from . import command


def setAngle(pin, angle):
    """Set angle of servo motor."""

    command(f"Setting angle {angle}", "sw", "SW-OK", bytes([pin, angle]))


def getAngle(pin):
    """Get angle of servo motor."""

    return (command("Getting angle", "sr", int, bytes([pin])))



"""CLI for Arduino API wrapper.
"""

import click
import sys

from . import ArduinoException, io, oled, rgb_lcd, shell


class ArduinoGroup(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except ArduinoException as e:
            sys.stderr.write(f"{e}\n")
            sys.exit(1)


@click.group(cls=ArduinoGroup)
def cli():
    # Load environment variables from .env if they exist
    import dotenv

    dotenv.load_dotenv()

    # Set up tracing, if desired
    import os

    try:
        from . import tracing
        tracing.set_default(os.getenv("TRACE"))
    except Exception as e:
        sys.stderr.write(f"{e}\n")
        sys.exit(1)


@cli.command()
@click.option("-d", "--device", default="grove")
def test(device):
    import time

    if device == "grove":
        oled.clear()
        print_message = oled.print_message
    else:
        rgb_lcd.clear()
        print_message = rgb_lcd.print

    print_message("Testing...")
    time.sleep(0.5)

    count = 0

    while True:
        print_message(count)
        io.digital_write(4, count%2)

        time.sleep(0.1)
        count += 1


@cli.group()
def analog():
    """Commands to interact with analog pins."""
    pass


@analog.command("read")
@click.option("-c", "--continuous", is_flag=True)
@click.argument("pin", type=int)
def aread(pin, continuous):
    """Read a value from an analog pin."""

    if continuous:
        while True:
            print(io.analog_read(pin))
    else:
        print(io.analog_read(pin))


@analog.command("write")
@click.argument("pin", type=int)
@click.argument("duty", type=int)
def awrite(pin, duty):
    """Write a duty cycle to an analog pin."""

    io.analog_write(pin, duty)


@cli.group()
def digital():
    """Commands to interact with digital pins."""
    pass


@digital.command("read")
@click.option("-c", "--continuous", is_flag=True)
@click.argument("pin", type=int)
def dread(pin, continuous):
    """Read a value from a digital pin."""

    if continuous:
        while True:
            print(io.digital_read(pin))
    else:
        print(io.digital_read(pin))


@digital.command("write")
@click.argument("pin", type=int)
@click.argument("value", type=bool)
def dwrite(pin, value):
    """Write a bit cycle to a digital pin."""

    io.digital_write(pin, value)


@cli.group("lcd")
def lcd_group():
    """Commands to interact with the RGB LCD."""
    pass


@lcd_group.command("clear")
def lcd_clear():
    rgb_lcd.clear()


@lcd_group.command("colour")
@click.argument("red", type=int)
@click.argument("green", type=int)
@click.argument("blue", type=int)
def lcd_colour(red, green, blue):
    rgb_lcd.set_colour(red, green, blue)


@lcd_group.command("print")
@click.option("-r", "--row", type=int, default=0)
@click.option("-c", "--col", type=int, default=0)
@click.argument("message")
def lcd_print(row, col, message):
    rgb_lcd.print(row, col, message)


@cli.group("oled")
def oled_group():
    """Commands to print to the OLED display."""
    pass


@oled_group.command()
def clear():
    """Clear the OLED display."""

    oled.clear()


@oled_group.command("print")
@click.argument("message", type=str)
def oled_print(message):
    """Print a message to the OLED display."""

    oled.clear()
    oled.print_message(message)


@cli.command("shell")
def run_shell():
    """Directly interact with the remote device."""

    from . import tracing
    tracing.set_default('pretty')

    shell.GroveShell().cmdloop()



def encode(v):
    """Generate TLV encoding of a Python value."""

    if type(v) == bool:
        ty = b"B"
        raw = int(v).to_bytes(1, byteorder="little")
    elif type(v) == int:
        ty = b"I"
        raw = v.to_bytes(2, byteorder="little")
    elif type(v) == str:
        ty = b"S"
        raw = v.encode("UTF-8")

    encoded = bytearray(ty)
    encoded.append(len(raw))
    encoded.extend(raw)

    return encoded


def decode(ty, raw_bytes):
    """Parse TLV-encoded bytes."""

    if ty == "B":
        if raw_bytes[0] == 0:
            return False
        elif raw_bytes[0] == 1:
            return True
        else:
            from . import ProtocolError
            raise ProtocolError(f"Invalid byte value: {raw_bytes}")

    elif ty == "E":
        from . import DeviceError
        raise DeviceError(raw_bytes.decode("ascii", "backslashreplace"))

    elif ty == "I":
        return int.from_bytes(raw_bytes, byteorder="little")

    elif ty == "S":
        return raw_bytes.decode("ascii")

    assert False and "unreachable"



import re
import serial

from . import tlv, tracing

BANNER = b"ENGI 1020 v"


class Link:
    """A serial link to a device that speaks the ENGI 1020 protocol."""

    # Assuming there's only one Grove kit connected to this computer, here it is.
    singleton = None

    @classmethod
    def get(cls, port=None, version=range(6)):
        if cls.singleton is None:
            if port is None:
                port = find_port()

            if port is None:
                from . import ArduinoException
                raise ArduinoException('Unable to find Arduino device')

            cls.singleton = Link(port, version)

        return cls.singleton

    def __init__(self, port, version_range, baud_rate=9600, timeout=5):
        from . import ProtocolError

        self.s = serial.Serial(
            port,
            baud_rate,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            timeout=timeout,
        )

        self.trace = tracing.default
        self.trace("Reading banner and version", action=True)

        banner = self.read_tlv()
        print(banner)
        m = re.match('ENGI 1020 v([0-9]+)(.*)', banner)
        if not m:
            raise ProtocolError(f"Invalid banner: '{banner}'")

        version, remainder = m.groups()

        self.version = int(version)
        if not self.version in version_range:
            raise ProtocolError(
                f"Unknown ENGI 1020 protocol version: {self.version}")

        if self.version >= 2:
            (self.board,) = re.match(' on (.*)', remainder).groups()
            self.signature = None
        if self.version >= 4:
            (self.board, self.signature) = (
                re.match('(.*) \((.*)\)', self.board).groups()
            )
        else:
            self.board = 'ENGI 1020'
            self.signature = None

    def command(self, descrip, command, expect, *params):
        """Execute a remote command.

        Parameters:
          descrip (str):         Human-readable description for tracing
          command (str|bytes):   Command identifier
          expect:                Expected response value, type or predicate
          params:                Byte string(s) to be sent as parameters
        """
        self.trace(descrip, action=True)

        self.write(command.encode() if type(command) == str else command)

        for param in params:
            self.write(param)

        self.trace("Reading response", action=True)
        r = self.read_tlv()

        if isinstance(expect, type(lambda x: x)):
            response_ok = expect(r)
        elif isinstance(expect, type):
            response_ok = isinstance(r, expect)
        else:
            response_ok = (r == expect)

        if not response_ok:
            from . import ProtocolError

            raise ProtocolError(
                f'Response to {command} command should be {expect}, not {r}')

        return r

    def read(self, size=1):
        self.trace("Reading", size, "B")

        b = self.s.read(size)

        if len(b) < size:
            from . import ProtocolError

            raise ProtocolError(f"Failed to read {size} B from {self.s}: {b}")

        self.trace("Read", len(b), "B:", b)

        return b

    def read_tlv(self):
        t = self.read(1)[0]
        len = self.read(1)[0]
        self.trace("Value of type", t, "/", chr(t), "has len", len)

        raw = self.read(len)

        return tlv.decode(chr(t), raw)

    def read_until(self, expected):
        self.trace(f"Reading until:", expected)

        b = self.s.read_until(expected)

        self.trace("Read", len(b), "B:", b)

        return b

    def write(self, raw_bytes):
        self.trace("Writing", len(raw_bytes), "B:", raw_bytes)
        return self.s.write(raw_bytes)

    def write_byte(self, b):
        self.write_integer(b, 1)

    def write_integer(self, n, bytes):
        self.write(n.to_bytes(bytes, byteorder="little"))


def find_port():
    """Find the serial device the Arduino device is attached to.

    With luck, this will be the only serial device. If we're not so lucky,

    If only one Arduino device is connected to the computer, leaving the default
    value of port=None will allow for auto-detection. However, the port value
    can be set explicitly to specify one of several options."""

    import serial.tools.list_ports
    import sys

    if sys.platform == "darwin":
        ports = list(serial.tools.list_ports.grep("cu.usb"))

    elif sys.platform == "win32":
        ports = list(serial.tools.list_ports.comports())

        if len(ports) != 1:
            print("Check Windows Device Manager for COM port number")

    else:
        ports = list(serial.tools.list_ports.grep("ttyUSB[0-9]+"))

        if len(ports) == 0:
            ports = list(serial.tools.list_ports.grep("ttyACM[0-9]+"))

    if len(ports) == 0:
        sys.stderr.write("No Arduino device detected (none plugged in?)\n")
        return None

    if len(ports) == 1:
        return ports[0].device

    print("Multiple devices detected:")
    for i, p in enumerate(ports):
        desc = f"({p.description})" if p.description != "n/a" else ""
        print(f"[{i}]    {p.name} {desc}")

    while True:
        try:
            i = int(input("select device> "))
            return ports[int(i)].device

        except IndexError as e:
            sys.stderr.write("Invalid device number\n")

        except ValueError as e:
            print(e)



"""Arduino API wrapper.
"""

from . import serial, tlv, tracing


class ArduinoException(Exception):
    """Base class for Arduino-related errors."""
    pass


class DeviceError(ArduinoException):
    """An error reported by the Arduino device."""
    pass


class ProtocolError(ArduinoException):
    """An error in the remote Arduino protocol."""
    pass


def command(*args, **kwargs):
    """Execute a remote Arduino command.

    See Link.command() for parameter details.
    """

    return serial.Link.get().command(*args, **kwargs)


def protocol_version():
    """What version of the serial protocol are we speaking?"""

    return serial.Link.get().version



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

setup(
    name="engi1020",
    version="5.1.1",
    description="Software library for Engineering 1020: Introduction to Programming at Memorial University.",
    license="MIT",
    author="Jonathan Anderson, Lori Hogan",
    packages=find_packages(),
    entry_points='''
        [console_scripts]
        arduino=engi1020.arduino.cli:cli
    ''',
    install_requires=[
        'click',
        'hexdump',
        'matplotlib',
        'pyserial',
        'python-dotenv',
    ],
    long_description=open("README.rst").read(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
)



from . import command


def getTemperature():
    """Get temperature from Grove Beginner Air Pressure (BMP280)."""

    return (command("Getting temperature (BMP280)", "pt", int)/100)

def getPressure():
    """Get air pressure (BMP280)."""

    return (command("Getting air pressure", "pp", int)/100)

def getAltitude():
    """Get altitude (BMP280)."""

    return (command("Getting altitude", "pa", int)/100)



from . import command


def getTemp(pin):
    """Get temperature reading from DHT11."""

    return (command("Getting temperature", "Dt", int, bytes([pin])))/100


def getHumidity(pin):
    """Get humidity reading from DHT11."""

    return (command("Getting humidity", "Dh", int, bytes([pin])))/100



import cmd
import shlex
import sys

from . import serial

class GroveShell(cmd.Cmd):
    intro = 'ENGI 1020 Grove shell\n'
    prompt = '> '

    def __init__(self):
        super().__init__()

        self.link = serial.Link.get()

        if self.link.signature:
            self.intro += f'Device {self.link.signature} ({self.link.board})\n'

        self.prompt = f'{self.link.board} v{self.link.version}> '

    def precmd(self, line):
        """Precondition a command line, parsing arguments into Python values.

        Returns: None, 'EOF' or (command, args) tuple.
        """

        cmdline = shlex.split(line)
        if len(cmdline) == 0:
            return None

        cmd = cmdline[0]
        if cmd == 'EOF':
            return 'EOF'

        try:
            args = [parse_arg(a) for a in cmdline[1:]]
            return (cmd, args)
        except Exception as e:
            sys.stderr.write(f"{e}\n")
            return None

    def onecmd(self, line):
        """Execute a single command."""

        if line is None:
            return

        if line == 'EOF' or line == 'exit':
            return True

        try:
            cmd, args = line
            serialized = [serialize(a) for a in args]
            msg = f"Executing {cmd} {' '.join([str(a) for a in args])}"

            print(self.link.command(msg, cmd, valid_response, *serialized))
        except Exception as e:
            sys.stderr.write(f"{e}\n")
            return None


def parse_arg(a):
    """Parse a single command-line argument."""

    try:
        return eval(a)
    except Exception as e:
        return a


def serialize(value):
    """Turn a value passed at the command line into serial bytes."""

    if isinstance(value, bool) or isinstance(value, int):
        return bytes([value])

    if isinstance(value, bytes):
        return value

    if isinstance(value, str):
        # Encode strings with length first
        encoded = value.encode('utf-8')
        b = bytearray([len(encoded)])
        b.extend(encoded)

        return bytes(b)


    raise ValueError(f"{value} has unknown type {type(value)}")


def valid_response(r):
    """Is a response from the Arduino device valid?"""

    return (r is not None)
